From: "Rafael J. Wysocki" <rjw@rjwysocki.net>
To: Yinghai Lu <yinghai@kernel.org>
Cc: Bjorn Helgaas <bhelgaas@google.com>,
"Rafael J. Wysocki" <rafael.j.wysocki@intel.com>,
Gu Zheng <guz.fnst@cn.fujitsu.com>,
Guo Chao <yan@linux.vnet.ibm.com>,
"linux-pci@vger.kernel.org" <linux-pci@vger.kernel.org>,
"linux-kernel@vger.kernel.org" <linux-kernel@vger.kernel.org>,
Mika Westerberg <mika.westerberg@linux.intel.com>
Subject: Re: [PATCH v2 04/10] PCI: Destroy pci dev only once
Date: Sat, 30 Nov 2013 00:38:26 +0100 [thread overview]
Message-ID: <461934992.nFRx8soJrp@vostro.rjw.lan> (raw)
In-Reply-To: <CAE9FiQXevSPmPEvxzXjC-CFhpYHic0v+vN5+JV3j5kmX-wv_GQ@mail.gmail.com>
On Tuesday, November 26, 2013 06:26:54 PM Yinghai Lu wrote:
> On Tue, Nov 26, 2013 at 5:24 PM, Rafael J. Wysocki <rjw@rjwysocki.net> wrote:
> >
> > So assume pci_destroy_dev() is called twice in parallel for the same dev
> > by two different threads. Thread 1 does the atomic_inc_and_test() and
> > finds that it is OK to do the device_del() and put_device() which causes
> > the device object to be freed. Then thread 2 does the atomic_inc_and_test()
> > on the already freed device object and crashes the kernel.
> >
> thread2 should still hold one extra reference.
> that is in
> device_schedule_callback
> ==> sysfs_schedule_callback
> ==> kobject_get(kobj)
>
> pci_destroy_dev for thread2 is called at this point.
>
> and that reference will be released from
> sysfs_schedule_callback
> ==> kobject_put()...
Well, that would be the case if thread 2 was started by device_schedule_callback(),
but again, for example, it may be trim_stale_devices() started by acpiphp_check_bridge()
that doesn't hold extra references to the pci_dev. [Well, that piece of code
is racy anyway, because it walks bus->devices without locking. Which is my
fault too, because I overlooked that. Shame, shame.]
Perhaps we can do something like the (untested) patch below (in addition to the
$subject patch). Do you see any immediate problems with it?
Also I wonder if it is safe to do pci_stop_and_remove_device() in acpiphp_glue.c
without putting it under pci_remove_rescan_mutex?
Rafael
---
drivers/pci/hotplug/acpiphp_glue.c | 47 +++++++++++++++++++++++++------------
1 file changed, 33 insertions(+), 14 deletions(-)
Index: linux-pm/drivers/pci/hotplug/acpiphp_glue.c
===================================================================
--- linux-pm.orig/drivers/pci/hotplug/acpiphp_glue.c
+++ linux-pm/drivers/pci/hotplug/acpiphp_glue.c
@@ -687,10 +687,11 @@ static unsigned int get_slot_status(stru
}
/**
* trim_stale_devices - remove PCI devices that are not responding.
* @dev: PCI device to start walking the hierarchy from.
+ * @trim_list: List of devices to remove.
*/
-static void trim_stale_devices(struct pci_dev *dev)
+static void find_stale_devices(struct pci_dev *dev, struct list_head *trim_list)
{
acpi_handle handle = ACPI_HANDLE(&dev->dev);
struct pci_bus *bus = dev->subordinate;
@@ -710,21 +711,46 @@ static void trim_stale_devices(struct pc
alive = pci_bus_read_dev_vendor_id(dev->bus, dev->devfn, &v, 0);
}
if (!alive) {
- pci_stop_and_remove_bus_device(dev);
- if (handle)
- acpiphp_bus_trim(handle);
+ pci_dev_get(dev);
+ list_move(&dev->bus_list, trim_list);
} else if (bus) {
struct pci_dev *child, *tmp;
/* The device is a bridge. so check the bus below it. */
pm_runtime_get_sync(&dev->dev);
list_for_each_entry_safe(child, tmp, &bus->devices, bus_list)
- trim_stale_devices(child);
+ find_stale_devices(child, trim_list);
pm_runtime_put(&dev->dev);
}
}
+void acpiphp_trim_stale_devices(struct acpiphp_slot *slot)
+{
+ struct pci_bus *bus = slot->bus;
+ struct pci_dev *dev, *tmp;
+ LIST_HEAD(trim_list);
+
+ down_write(&pci_bus_sem);
+
+ list_for_each_entry_safe(dev, tmp, &bus->devices, bus_list)
+ if (PCI_SLOT(dev->devfn) == slot->device)
+ find_stale_devices(dev, &trim_list);
+
+ up_write(&pci_bus_sem);
+
+ while (!list_empty(&trim_list)) {
+ acpi_handle handle;
+
+ dev = list_first_entry(&trim_list, struct pci_dev, bus_list);
+ handle = ACPI_HANDLE(&dev->dev);
+ pci_stop_and_remove_bus_device(dev);
+ pci_dev_put(dev);
+ if (handle)
+ acpiphp_bus_trim(handle);
+ }
+}
+
/**
* acpiphp_check_bridge - re-enumerate devices
* @bridge: where to begin re-enumeration
@@ -737,18 +763,11 @@ static void acpiphp_check_bridge(struct
struct acpiphp_slot *slot;
list_for_each_entry(slot, &bridge->slots, node) {
- struct pci_bus *bus = slot->bus;
- struct pci_dev *dev, *tmp;
-
mutex_lock(&slot->crit_sect);
/* wake up all functions */
if (get_slot_status(slot) == ACPI_STA_ALL) {
/* remove stale devices if any */
- list_for_each_entry_safe(dev, tmp, &bus->devices,
- bus_list)
- if (PCI_SLOT(dev->devfn) == slot->device)
- trim_stale_devices(dev);
-
+ acpiphp_trim_stale_devices(slot);
/* configure all functions */
enable_slot(slot);
} else {
next prev parent reply other threads:[~2013-11-29 23:25 UTC|newest]
Thread overview: 69+ messages / expand[flat|nested] mbox.gz Atom feed top
2013-11-26 1:28 [PATCH v2 00/10] PCI: Double removing fix and allocate 64bit mmio pref Yinghai Lu
2013-11-26 1:28 ` [PATCH v2 01/10] PCI: Use device_release_driver in pci_stop_root_bus Yinghai Lu
2013-11-27 1:09 ` Rafael J. Wysocki
2013-11-26 1:28 ` [PATCH v2 02/10] PCI: Move back pci_proc_attach_devices calling Yinghai Lu
2013-11-26 1:28 ` [PATCH v2 03/10] PCI: Move resources and bus_list releasing to pci_release_dev Yinghai Lu
2013-11-27 1:15 ` Rafael J. Wysocki
2013-11-27 2:15 ` Yinghai Lu
2013-11-26 1:28 ` [PATCH v2 04/10] PCI: Destroy pci dev only once Yinghai Lu
2013-11-26 3:38 ` Bjorn Helgaas
2013-11-26 19:34 ` Yinghai Lu
2013-11-26 20:13 ` Yinghai Lu
2013-11-27 1:24 ` Rafael J. Wysocki
2013-11-27 2:26 ` Yinghai Lu
2013-11-29 23:38 ` Rafael J. Wysocki [this message]
2013-11-29 23:45 ` Rafael J. Wysocki
2013-11-30 0:31 ` Rafael J. Wysocki
2013-11-30 21:37 ` Rafael J. Wysocki
2013-11-30 22:27 ` Yinghai Lu
2013-12-01 1:24 ` Rafael J. Wysocki
2013-12-02 1:29 ` Rafael J. Wysocki
2013-12-02 14:49 ` Rafael J. Wysocki
2013-12-05 22:40 ` Bjorn Helgaas
2013-12-06 1:21 ` Rafael J. Wysocki
2013-12-06 6:29 ` Yinghai Lu
2014-01-10 14:20 ` [PATCH 0/9] PCI: Eliminate race conditions between hotplug and sysfs rescan/remove (Was: Re: [PATCH v2 04/10] PCI: Destroy pci dev only once) Rafael J. Wysocki
2014-01-10 14:22 ` [PATCH 1/9] PCI: Global rescan-remove lock Rafael J. Wysocki
2014-01-10 14:23 ` [PATCH 2/9] ACPI / PCI: Use global PCI rescan-remove locking in PCI root hotplug Rafael J. Wysocki
2014-01-10 14:24 ` [PATCH 3/9] ACPI / hotplug / PCI: Use global PCI rescan-remove locking Rafael J. Wysocki
2014-01-10 14:25 ` [PATCH 4/9] PCMCIA / cardbus: " Rafael J. Wysocki
2014-01-10 14:26 ` [PATCH 5/9] PCI / hotplug: " Rafael J. Wysocki
2014-01-10 14:27 ` [PATCH 6/9] platform / x86: " Rafael J. Wysocki
2014-01-10 14:27 ` [PATCH 7/9] MPT / PCI: Use pci_stop_and_remove_bus_device_locked() Rafael J. Wysocki
2014-01-10 14:28 ` [PATCH 8/9] powerpc / eeh_driver: Use global PCI rescan-remove locking Rafael J. Wysocki
2014-01-15 13:36 ` [Update][PATCH " Rafael J. Wysocki
2014-01-15 17:38 ` Bjorn Helgaas
2014-01-10 14:29 ` [PATCH 9/9] Xen / PCI: " Rafael J. Wysocki
2014-01-15 18:02 ` [PATCH 0/9] PCI: Eliminate race conditions between hotplug and sysfs rescan/remove (Was: Re: [PATCH v2 04/10] PCI: Destroy pci dev only once) Bjorn Helgaas
2013-12-06 6:52 ` [PATCH v2 04/10] PCI: Destroy pci dev only once Yinghai Lu
2013-12-07 1:27 ` Rafael J. Wysocki
2013-12-08 3:31 ` Yinghai Lu
2013-12-08 3:50 ` Greg Kroah-Hartman
2013-12-09 15:24 ` Ethan Zhao
2013-12-09 19:08 ` Greg Kroah-Hartman
2013-12-10 7:43 ` Ethan Zhao
2014-01-13 1:03 ` [PATCH] PCI / remove: Check parent kobject in pci_destroy_dev() (was: Re: [PATCH v2 04/10] PCI: Destroy pci dev only once) Rafael J. Wysocki
2013-11-27 1:17 ` [PATCH v2 04/10] PCI: Destroy pci dev only once Rafael J. Wysocki
2013-11-26 1:28 ` [PATCH v2 05/10] PCI: pcibus address to resource converting take bus directly Yinghai Lu
2013-11-26 1:28 ` [PATCH v2 06/10] PCI: Add pcibios_bus_addr_to_res() Yinghai Lu
2013-11-26 1:28 ` [PATCH v2 07/10] PCI: Try to allocate mem64 above 4G at first Yinghai Lu
2013-11-26 4:15 ` Bjorn Helgaas
2013-11-26 20:14 ` Yinghai Lu
2013-11-26 1:28 ` [PATCH v2 08/10] PCI: Try best to allocate pref mmio 64bit above 4g Yinghai Lu
2013-11-26 4:17 ` Bjorn Helgaas
2013-11-26 6:59 ` Guo Chao
2013-11-26 17:53 ` Bjorn Helgaas
2013-11-26 22:00 ` Yinghai Lu
2013-11-26 22:01 ` Bjorn Helgaas
2013-11-27 0:33 ` Yinghai Lu
2013-11-26 1:28 ` [PATCH v2 09/10] PCI: Sort pci root bus resources list Yinghai Lu
2013-11-26 4:18 ` Bjorn Helgaas
2013-11-26 1:28 ` [PATCH v2 10/10] intel-gtt: Read 64bit for gmar_bus_addr Yinghai Lu
2013-11-26 3:46 ` Bjorn Helgaas
2013-11-26 19:35 ` Yinghai Lu
2013-12-11 18:48 ` Bjorn Helgaas
2013-12-11 19:58 ` Yinghai Lu
2013-12-21 0:27 ` Bjorn Helgaas
2013-12-21 1:19 ` Yinghai Lu
2013-12-21 18:50 ` Bjorn Helgaas
2013-12-23 22:33 ` Bjorn Helgaas
Reply instructions:
You may reply publicly to this message via plain-text email
using any one of the following methods:
* Save the following mbox file, import it into your mail client,
and reply-to-all from there: mbox
Avoid top-posting and favor interleaved quoting:
https://en.wikipedia.org/wiki/Posting_style#Interleaved_style
* Reply using the --to, --cc, and --in-reply-to
switches of git-send-email(1):
git send-email \
--in-reply-to=461934992.nFRx8soJrp@vostro.rjw.lan \
--to=rjw@rjwysocki.net \
--cc=bhelgaas@google.com \
--cc=guz.fnst@cn.fujitsu.com \
--cc=linux-kernel@vger.kernel.org \
--cc=linux-pci@vger.kernel.org \
--cc=mika.westerberg@linux.intel.com \
--cc=rafael.j.wysocki@intel.com \
--cc=yan@linux.vnet.ibm.com \
--cc=yinghai@kernel.org \
/path/to/YOUR_REPLY
https://kernel.org/pub/software/scm/git/docs/git-send-email.html
* If your mail client supports setting the In-Reply-To header
@@ -687,10 +687,11 @@ static unsigned int get_slot_status(stru
}
/**
- * trim_stale_devices - remove PCI devices that are not responding.
+ * find_stale_devices - Select PCI devices that are not responding for removal.
* @dev: PCI device to start walking the hierarchy from.
via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line
before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox
all inboxes | Powered by JetHome®